import builtins
import json
from typing import Type, Optional, TypeVar

import click

from .cli import gameta_cli
from .context import gameta_context, GametaContext


__all__ = ['parameters_cli']


T = TypeVar('T', int, float, str, bool, dict, list)


@gameta_cli.group('params')
@gameta_context
def parameters_cli(context: GametaContext) -> None:
    """
    CLI for managing parameters in metarepos
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None
    """
    if not context.is_metarepo:
        raise click.ClickException(f"{context.project_dir} is not a metarepo, initialise it with 'gameta init'")


@parameters_cli.command()
@click.option('--param', '-p', type=str, required=True, help="Parameter name to be added to all child repositories")
@click.option('--type', '-t', 'ptype', type=click.Choice(['int', 'float', 'str', 'bool', 'dict', 'list']),
              default='str', help='Parameter type to be added')
@click.option('--value', '-v', type=str, default=None, help="Default value to be used for all child repositories")
@click.option('--skip-prompt', '-y', 'skip_prompt', is_flag=True, default=False,
              help="Skip user prompt and use default values for all")
@gameta_context
def add(context: GametaContext, param: str, ptype: Optional[str], value: Optional[str], skip_prompt: bool) -> None:
    """
    Adds/updates a parameter to all child repositories in the .meta file, users will automatically be prompted to enter
    the parameter value corresponding to each repository. They can provide a type and value parameter to skip this step.
    \f
    Args:
        context (GametaContext): Gameta Context
        param (str): Parameter name to be added to all child repositories
        ptype (Optional[str]): Parameter type to be added
        value (Optional[str]): Default value to be used for all child repositories, value will be rendered with type
                               chosen
        skip_prompt (bool): Flag to indicate if user prompt to enter values for all repositories should be skipped and
                            default value used directly

    Returns:
        None

    Examples:
        $ gameta parameters add -p test -t str -v hello_world  # Adds parameter test with value hello_world to all repos
        $ gameta parameters add -p test -t str -v hello_world -u  # Prompts user input for parameter values

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Adding parameter {param}")
    parameter_type: Type[T] = getattr(builtins, ptype, str)

    def prompt_type_parser(input_value: str) -> T:
        """
        Parser to convert the user inputs into the appropriate type
        
        Args:
            input_value (str): Input received from the command line

        Returns:
            (int, float, str, bool, list, dict): Output parsed into the correct type
        """
        try:
            return json.loads(input_value)
        except json.JSONDecodeError:
            return parameter_type(input_value)

    try:
        for repo, details in context.repositories.items():
            if skip_prompt:
                pvalue: parameter_type = value
            else:
                pvalue: parameter_type = click.prompt(
                    f"Please enter the parameter value for repository {repo} or >* to skip", default=value,
                    value_proc=prompt_type_parser
                )
                if pvalue == ">*":
                    click.echo(f"Skip token was entered, defaulting to default value {value}")
                    pvalue = value
                elif not isinstance(pvalue, parameter_type):
                    click.echo(f"Value {pvalue} (type: {type(pvalue)}) entered is not the required type "
                               f"{parameter_type}, defaulting to default value {value}")
                    pvalue = value
            details[param] = pvalue
            click.echo(f"Adding {param} value {pvalue} for {repo}")
        context.export()
        click.echo(f"Successfully added parameter {param} to .meta file")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@parameters_cli.command()
@click.option('--param', '-p', type=str, required=True,
              help="Parameter name to be deleted from all child repositories")
@gameta_context
def delete(context: GametaContext, param: str) -> None:
    """
    Deletes a parameter from all child repositories
    \f
    Args:
        context (GametaContext): Gameta Context
        param (str): Parameter name to be deleted from all child repositories

    Returns:
        None

    Raises:
        click.ClickException: If errors occur during processing
    """
    if param in context.reserved_params:
        raise click.ClickException(f"Parameter {param} is a reserved parameter {context.reserved_params}")

    try:
        click.echo(f"Deleting parameter {param}")
        for repo, details in context.repositories.items():
            try:
                del details[param]
            except KeyError:
                continue
        context.export()
        click.echo(f"Successfully deleted parameter {param} from .meta file")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")



import json
import shlex
from contextlib import contextmanager
from copy import deepcopy
from os import getenv, getcwd, chdir
from os.path import join, basename, normpath, abspath
from typing import Optional, List, Generator, Dict, Tuple

import click


__all__ = [
    # Contexts
    'GametaContext', 'gameta_context',
]


SHELL = getenv('SHELL', '/bin/sh')


class GametaContext(object):
    """
    GametaContext for the current Gameta session

    Attributes:
        project_dir (Optional[str]): Project directory
        is_metarepo (bool): Project is a metarepo
        gameta_data (Dict): Gameta data extracted and exported
        repositories (Dict[str, Dict]): Data of all the repositories contained in the metarepo
        tags (Dict[str, List[str]]): Repository data organised according to tags
    """
    reserved_params = ['url', 'path', 'tags', '__metarepo__']

    def __init__(self):
        self.is_metarepo: bool = False
        self.gitignore_data: List[str] = []
        self.project_dir: Optional[str] = None
        self.gameta_data: Dict = {}
        self.commands: Dict = {}
        self.repositories: Dict[str, Dict] = {}
        self.tags: Dict[str, List[str]] = {}

    @property
    def project_name(self) -> str:
        """
        Returns the name of the project

        Returns:
            str: Name of the project
        """
        return basename(self.project_dir)

    @property
    def meta(self) -> str:
        """
        Returns the path to the .meta file of the project, i.e. where it should be if the Project has not been
        initialised

        Returns:
            str: Path to the project's .meta file
        """
        return join(self.project_dir, '.meta')

    @property
    def gitignore(self) -> str:
        """
        Returns the path to the .gitignore file of the project, i.e. where it should be if the Project has not been
        initialised

        Returns:
            str: Path to the project's .gitignore file
        """
        return join(self.project_dir, '.gitignore')

    def add_gitignore(self, path: str) -> None:
        """
        Adds the path to the gitignore_data

        Args:
            path (str): Path to be added

        Returns:
            None
        """
        self.gitignore_data.append(path + '/\n')

    def remove_gitignore(self, path: str) -> None:
        """
        Removes the path from the gitignore_data

        Args:
            path (str): Path to be removed

        Returns:
            None
        """
        try:
            self.gitignore_data.remove(path + '/\n')
        except ValueError:
            return

    def is_primary_metarepo(self, repo: str) -> bool:
        """
        Returns a boolean if the repository is a primary meta-repository

        Args:
            repo (str): Repository to check

        Returns:
            bool: Flag to indicate if repository is a primary meta-repository
        """
        return abspath(self.repositories[repo]["path"]) == self.project_dir

    def load(self) -> None:
        """
        Finds all repositories to manage and groups them into relative groups

        Returns:
            None
        """
        try:
            with open(self.meta, 'r') as f:
                self.gameta_data = json.load(f)
                self.repositories = self.gameta_data['projects']
                self.commands = self.gameta_data.get('commands', {})
            self.is_metarepo = True
        except FileNotFoundError:
            return
        except Exception as e:
            self.repositories = {}
            self.commands = {}
            click.echo(f"Could not load .meta file due to: {e.__class__.__name__}.{str(e)}")
            return

        try:
            self.generate_tags()
        except Exception as e:
            self.repositories = {}
            self.commands = {}
            click.echo(f"Malformed .meta file, error: {e.__class__.__name__}.{str(e)}")
            return

        try:
            with open(self.gitignore, 'r') as f:
                self.gitignore_data = f.readlines()
        except FileNotFoundError:
            return
        except Exception as e:
            self.gitignore_data = []
            click.ClickException(f"Could not load .gitignore file due to: {e.__class__.__name__}.{str(e)}")
            return

    def export(self) -> None:
        """
        Exports updated Gameta data to .meta file and gitignore data to the .gitignore file

        Returns:
            None
        """
        try:
            self.gameta_data['projects'] = self.repositories
            self.gameta_data['commands'] = self.commands
            with open(self.meta, 'w+') as f:
                json.dump(self.gameta_data, f)
        except Exception as e:
            raise click.ClickException(
                f"Could not export gameta data to .meta file: {e.__class__.__name__}.{str(e)}"
            )

        try:
            with open(self.gitignore, 'w+') as f:
                f.writelines(self.gitignore_data)
        except Exception as e:
            raise click.ClickException(
                f"Could not export gitignore data to .gitignore file: {e.__class__.__name__}.{str(e)}"
            )

    def generate_tags(self) -> None:
        """
        Updates the tag indexes of the repositories

        Returns:
            None
        """
        for repo, details in self.repositories.items():
            for tag in details.get('tags', []):
                if tag in self.tags:
                    self.tags[tag].append(repo)
                else:
                    self.tags[tag] = [repo]

    def apply(
            self,
            commands: List[str],
            repos: List[str] = (),
            shell: bool = False
    ) -> Generator[Tuple[str, str], None, None]:
        """
        Yields a list of commands to all repositories or a selected set of them, substitutes relevant parameters stored
        in .meta file

        Args:
            commands (List[str]): Commands to be applied
            repos (List[str]): Selected set of repositories
            shell (bool): Flag to indicate if a separate shell should be used

        Returns:
            None
        """
        repositories: List[Tuple[str, Dict[str, str]]] = \
            [(repo, details) for repo, details in self.repositories.items() if repo in repos] or \
            list(self.repositories.items())

        for repo, details in repositories:
            with self.cd(details['path']):
                repo_commands: List[str] = [c.format(**details) for c in deepcopy(commands)]
                command: List[str] = self.shell(repo_commands) if shell else self.tokenise(' && '.join(repo_commands))
                yield repo, command

    @staticmethod
    def tokenise(command: str) -> List[str]:
        return shlex.split(command)

    @contextmanager
    def cd(self, sub_directory: str) -> Generator[str, None, None]:
        """
        Changes directory to a subdirectory within the project

        Args:
            sub_directory (str): Relative subdirectory within the project

        Returns:
            Generator[str, None, None]: Path to current directory
        """
        cwd = getcwd()
        path = normpath(join(self.project_dir, sub_directory.lstrip('/')))
        chdir(path)
        yield path
        chdir(cwd)

    def shell(self, commands: List[str]) -> List[str]:
        """
        Executes commands provided in a separate shell as subprocess does not natively handle piping

        Args:
            commands (List[str]): User-defined commands

        Returns:
            List[str]: Shell command string to be executed by subprocess
        """
        return self.tokenise(
            f'{SHELL} -c " ' +
            ' && '.join(commands) +
            '"'
        )


gameta_context = click.make_pass_decorator(GametaContext, ensure=True)
